#! /usr/bin/env python
#encoding:UTF-8

'''
Created on 03/05/2010

This program allows you to calibrate a camera through a set of images (chessboards). 
Images are stored in a directory. 
The PATH variable contains the directory path and type of image.

The chessboard_test.zip file contains a set images for test.

@author: jbekios
'''

import glob
import cv2.cv as cv

PATH_ORIGIN = "chessboard/*.jpg"

list_files = glob.glob(PATH_ORIGIN)
list_files.sort()

# Create a structure for a gray image
rgb_image = cv.LoadImage(list_files[0], cv.CV_LOAD_IMAGE_COLOR)
gray_image = cv.CreateImage(cv.GetSize(rgb_image), cv.IPL_DEPTH_8U, 1)

# Window to show images
cv.NamedWindow("Show image", 0)
cv.NamedWindow("Show corners", 0)

# Size chessboard
chessboard_dim = (8,5) #width, height

# Get the image data
n_boards = len(list_files) # Number of boards
print " Number of boards", n_boards

board_w = chessboard_dim[0]
board_h = chessboard_dim[1]
board_n = board_w * board_h # Data by chessboard

# Save the image information and objects in the world
image_points = cv.CreateMat(n_boards*board_n, 2, cv.CV_32FC1)
object_points = cv.CreateMat(n_boards*board_n, 3, cv.CV_32FC1)
points_counts = cv.CreateMat(n_boards, 1, cv.CV_32SC1)
intrinsic_matrix = cv.CreateMat(3, 3, cv.CV_32FC1)
distortion_coeffs = cv.CreateMat(4, 1, cv.CV_32FC1)

successes = 0
total_corners = 0
# Read images of a directory
for file_image in list_files:
    print "Picture: ", file_image
    # Load image
    rgb_image = cv.LoadImage(file_image, cv.CV_LOAD_IMAGE_COLOR)
    cv.ShowImage("Show picture", rgb_image)
    # Corners detected (It works best if the image is in color)
    found_all, corners = cv.FindChessboardCorners(rgb_image, chessboard_dim, cv.CV_CALIB_CB_ADAPTIVE_THRESH|cv.CV_CALIB_CB_NORMALIZE_IMAGE)
    # Improving detection of corners
    print "Corners: ", len(corners)
    cv.CvtColor(rgb_image, gray_image, cv.CV_RGB2GRAY)
    corners = cv.FindCornerSubPix(gray_image, corners, (2,2), (-1,-1), (cv.CV_TERMCRIT_EPS+cv.CV_TERMCRIT_ITER, 30, 0.1))
    # Draw corners
    cv.DrawChessboardCorners(rgb_image, chessboard_dim, corners, found_all)
    cv.ShowImage("Show Corners", rgb_image)
    
    # If we got a good board, add it to our data
    counter_count = len(corners)
    print "Counter count: ", counter_count
    if (counter_count == board_n):
        step = successes * board_n
        print "step: ", step
        i = step
        j = 0
        while j < board_n:
            # Add the data to the matrix
            image_points[i, 0] = corners[j][0]
            image_points[i, 1] = corners[j][1]
            object_points[i, 0] = j//board_w
            object_points[i, 1] = j%board_w
            object_points[i, 2] = 0.0
            j += 1
            i += 1
        points_counts[successes, 0] = board_n
        total_corners += board_n
        successes += 1           
    else:
        break
        
    cv.WaitKey()

# CALIBRATE THE CAMERA!
cv.SetZero(intrinsic_matrix)
cv.SetZero(distortion_coeffs)
intrinsic_matrix[0, 0] = 1.0
intrinsic_matrix[1, 1] = 1.0

print "n_board", n_boards
print "total_corners", total_corners

rvects = cv.CreateMat(n_boards, 3, cv.CV_32FC1)
tvects = cv.CreateMat(n_boards, 3, cv.CV_32FC1)
cv.CalibrateCamera2(object_points, 
                    image_points, 
                    points_counts, 
                    cv.GetSize(gray_image), 
                    intrinsic_matrix, 
                    distortion_coeffs,           
                    rvects, 
                    tvects,
                    0)

print
print "Intrinsic matrix"

for i in xrange(3):
    print [intrinsic_matrix[i,j] for j in xrange(3)]

print "Vector rotation"
rotRod = cv.CreateMat(1, 3, cv.CV_32FC1)
matRot = cv.CreateMat(3, 3, cv.CV_32FC1)

for i in xrange(n_boards):
    rotRod = rvects[i,:]
    cv.Rodrigues2(rotRod, matRot)
    print "IMAGE: %3.3d" % i
    for row in xrange(3):
        print [matRot[row, col] for col in xrange(3)]
