#! /usr/bin/env python
print "OpenCV Python version of lkdemo and facedetect"
import cv2.cv as cv
import sys

#############################################################################
# "user constants"

# webcam size
# size = (640, 480)
size = (320, 240)
# size = (1280, 960)

#####################################################
####### some "global" variables for lkdemo ##########
#####################################################
image = None
pt = None
add_remove_pt = False
flags = 0
night_mode = False

####################################################
##### some "global" variables for facedetect #######
####################################################
cascade = cv.Load('haarcascade_frontalface_alt.xml')
min_size = (20, 20)
image_scale = 3
haar_scale = 1.2
min_neighbors = 2
haar_flags = 0
win_size = 10
MAX_COUNT = 500

####################################################
############ the mouse callback  ###################
####################################################

# the callback on the trackbar
def on_mouse (event, x, y, flags, param):
    # we will use the global pt and add_remove_pt
    global pt
    global add_remove_pt
    if image is None:
        # not initialized, so skip
        return
    if image.origin != 0:
        # different origin
        y = image.height - y
    if event == cv.CV_EVENT_LBUTTONDOWN:
        # user has click, so memorize it
        pt = (x, y)
        print "Point clicked:", pt
        add_remove_pt = True

def find_facedetection_points(img):
    t = cv.GetTickCount()
    pt1, pt2 = facedetect(img)
    t = cv.GetTickCount() - t
    print "Face detection time = %gms" % (t/(cv.GetTickFrequency()*1000.))
    print "Face detected frame at:", pt1, pt2
    if pt1 != -1 and pt2 != -1:
        fdx = pt2[0]-pt1[0]
        fdy = pt2[1]-pt1[1]
        pt11 = [pt1[0]+20, pt1[1]+20]
        pt22 = [pt2[0]-20, pt2[1]-20]
        points = [(pt11[0], pt11[1]), (pt22[0], pt11[1]), (pt11[0]+(pt22[0]-pt11[0])/2, pt11[1]+(pt22[1]-pt11[1])/2), (pt11[0], pt22[1]), (pt22[0], pt22[1])]
        return points, pt1, pt2, fdx, fdy
    else:
        return [], -1, -1, -1, -1

def facedetect(img, drawrectangle=False):
    # allocate temporary images
    gray = cv.CreateImage((img.width,img.height), 8, 1)
    small_img = cv.CreateImage((cv.Round(img.width / image_scale),
                cv.Round (img.height / image_scale)), 8, 1)
    # convert color input image to grayscale
    cv.CvtColor(img, gray, cv.CV_BGR2GRAY)
    # scale input image for faster processing
    cv.Resize(gray, small_img, cv.CV_INTER_LINEAR)
    cv.EqualizeHist(small_img, small_img)
    if(cascade):
        faces = cv.HaarDetectObjects(small_img, cascade, cv.CreateMemStorage(0),
                                    haar_scale, min_neighbors, haar_flags, min_size)
        if faces:
            for ((x, y, w, h), n) in faces:
                # the input to cv.HaarDetectObjects was resized, so scale the 
                # bounding box of each face and convert it to two CvPoints
                pt1 = (int(x * image_scale), int(y * image_scale))
                pt2 = (int((x + w) * image_scale), int((y + h) * image_scale))
                if drawrectangle:
                    cv.Rectangle(img, pt1, pt2, cv.RGB(255, 0, 0), 3, 8, 0)
            return pt1, pt2
        else:
            return -1, -1
    # cv.ShowImage("result", img)
    
def add_point(pt, grey, features):
    global win_size
    # refine this corner location and append it to 'features'
    features += cv.FindCornerSubPix (grey, [pt], (win_size, win_size),\
    (-1, -1), (cv.CV_TERMCRIT_ITER | cv.CV_TERMCRIT_EPS, 20, 0.03))
    return features

def fdcorrect(fdx, fdy, fdx_old, fdy_old):
    if fdx == -1 and fdy ==-1:
        fdx = fdx_old
        fdy = fdy_old
    else:
        fdx_old = fdx
        fdy_old = fdy
    return fdx, fdy, fdx_old, fdy_old
    

def ptptdist(p0, p1):
    dx = p0[0] - p1[0]
    dy = p0[1] - p1[1]
    return dx**2 + dy**2

#############################################################################
# so, here is the main part of the program

if __name__ == '__main__':
    features = [] # list of tracking points
    frames = []
    fdx = -1 # x distance head width init
    fdy = -1 # y distance head height init
    capture = cv.CaptureFromCAM(0) # init camera
    # set camera size
    del_size = (int(size[0]/8), int(size[1]/8))
    cv.SetCaptureProperty(capture, cv.CV_CAP_PROP_FRAME_WIDTH, size[0])
    cv.SetCaptureProperty(capture, cv.CV_CAP_PROP_FRAME_HEIGHT, size[1])
    # display a small howto use it
    print "Hot keys: \n" \
          "\tESC - quit the program\n" \
          "\tc - delete all the points\n" \
          "\tn - switch the \"night\" mode on/off\n" \
          "To add/remove a feature point click it\n"
    # first, create the necessary windows
    cv.NamedWindow ('LkDemo and Facedetect', cv.CV_WINDOW_AUTOSIZE)
    # register the mouse callback
    cv.SetMouseCallback ('LkDemo and Facedetect', on_mouse, None)
    # get the first camera frame
    frame = cv.QueryFrame(capture)
    # init the images for the first time
    if image is None:
        # create the images we need
        image = cv.CreateImage (cv.GetSize (frame), 8, 3)
        image.origin = frame.origin
        grey = cv.CreateImage (cv.GetSize (frame), 8, 1)
        prev_grey = cv.CreateImage (cv.GetSize (frame), 8, 1)
        pyramid = cv.CreateImage (cv.GetSize (frame), 8, 1)
        prev_pyramid = cv.CreateImage (cv.GetSize (frame), 8, 1)
    fdx_old = fdx # save old head width in case next is not found
    fdy_old = fdy # save old head height in case next is not found
    # do the face detection, get the boundary points and width and height of head
    points, pt1, pt2, fdx, fdy = find_facedetection_points(frame)
    # test if head was found: fdx != -1 and fdy != -1 
    # if not use the old values from fdx_old and fdy_old
    # if it was found, save the new width and height in fdx_old and fdy_old
    fdx, fdy, fdx_old, fdy_old = fdcorrect(fdx, fdy, fdx_old, fdy_old)
    # print "features", features
    fd = 0 # init frame counter
    avg = 0 # init average point of all tracking points
    avg_run = 0 # init running average point of all tracking points
    pc = 0 # point check index
    need_to_init = False # for resetting and rereading the tracking points
    # distances of min to max x of all tracking points in x axis
    ddx = 0
    # distances of min to max y of all tracking points in y axis
    ddy = 0
    # ddy_run_afterpoint = 0
    ddy_run = 0
    # the default parameters for the search of all tracking points
    quality = 0.01
    min_distance = 10
    corrx = int(size[0]/16)
    corry = int(corrx/4*3)
    t_avg = 0
    while True: # do until break by keyboard
        # benchmark testing time init
        t1 = cv.GetTickCount()
        # increase frame counter
        fd = fd + 1
        # after 20 frames do a new face detection
        if fd == 20 or avg == [0, 0]:
            # new facedetection
            fd = 0 # reset flag
            # do the face detection, get the boundary points and width and height of head
            points, pt1, pt2, fdx, fdy = find_facedetection_points(frame)
            print (pt1, pt2, fdx, fdy )
            # as before:
            fdx, fdy, fdx_old, fdy_old = fdcorrect(fdx, fdy, fdx_old, fdy_old)
            # if anything was found:
            if points != []:
                # features = []
                # delete the first point in the tracking point field features
                try:
                    features.pop(0)
                except:
                    pass
        # get the number of tracking points in features
        l = len(features)
        # if number of tracking points > 16 delete the first one
        if l>16:
            features.pop(0)
        elif l>0 and avg_run != 0:
            # check if point with index pc is out of bounds
            if pc>=l:
                pc = 0
            # print "Point", pc, abs(features[pc][0]-avg_run[0]), abs(features[pc][1]-avg_run[1])
            # delete tracking points which have lost track with the average and which
            # are too far away
            if abs(features[pc][0]-avg_run[0])>del_size[0] or abs(features[pc][1]-avg_run[1])>del_size[1]:
                features.pop(pc)
            pc = pc + 1
        # if the new field of points is not empty 
        # shift one point after the other into display
        # therefore the flag add_remove_pt is set to True
        # delete this point from the field
        if points != []:
            pt = points.pop()
            # print "points", points
            add_remove_pt = True
        # get new frame
        frame = cv.QueryFrame(capture)
        # copy the frame, so we can draw on it
        cv.Copy (frame, image)
        # if there are any tracking points get
        # the average point of all
        if l>0: 
            avg = [0, 0]
            for k in features:
                avg = [avg[0]+k[0], avg[1]+k[1]]
            avg = [int(avg[0]/l), int(avg[1]/l)]
            # first time value assignment of avg_run
            if avg_run == 0: 
                avg_run = avg
            elif avg == [0, 0]:
                avg_run = [int(size[0]/2), int(size[1]/2)]
            # or avg and avg_run are not identical and need approach to each other
            elif avg_run != avg:
                if avg[0] != avg_run[0]:
                    avg_x = avg[0]-avg_run[0]
                    avg_x_step = int(abs(avg_x/5))
                    print avg, avg_run, avg_x
                    avg_run[0] = avg_run[0] + avg_x/abs(avg_x)*avg_x_step
                if avg[1] != avg_run[1]:
                    avg_y = avg[1]-avg_run[1]
                    avg_y_step = int(abs(avg_y/5))
                    avg_run[1] = avg_run[1] + avg_y/abs(avg_y)*avg_y_step
            # print "avg", avg
        # if features != []:
        #    print features
        # e.g. features = [(303.60311889648437, 96.077102661132813), (268.115478515625, 136.42671203613281)]
        # create a grey version of the image
        cv.CvtColor (image, grey, cv.CV_BGR2GRAY)
        if night_mode:
            # night mode: only display the points
            cv.SetZero (image)
        facefound = (pt1 != -1 and fdx != -1 and fdy != -1)
        # print "facefound", facefound
        if need_to_init and facefound or fd == 0 and facefound:
            # we want to search all the good points
            # create the wanted images
            eig = cv.CreateImage (cv.GetSize (grey), 32, 1)
            temp = cv.CreateImage (cv.GetSize (grey), 32, 1)
            features_old = features[:]
            exception_cv = False
            target_rect = (pt1[0]+corrx, pt1[1]+corry, fdx-corrx*2, fdy-corry*2)
            print target_rect
            try:
                cv.SetImageROI(grey, target_rect)
                # cv.SetImageROI(eig, target_rect)
                # cv.SetImageROI(temp, target_rect)
                # search the good points
                features = cv.GoodFeaturesToTrack (
                    grey, eig, temp,
                    MAX_COUNT,
                    quality, min_distance, None, 3, 0, 0.04)
                # refine the corner locations
                if fdx-corrx*2>25:
                    features = cv.FindCornerSubPix (
                    grey,
                    features,
                    (win_size, win_size),  (-1, -1),
                    (cv.CV_TERMCRIT_ITER | cv.CV_TERMCRIT_EPS, 20, 0.03))
                else:
                    features = features_old
                    exception_cv = True
                cv.ResetImageROI(grey)
            except:
                exception_cv = True
            # cv.ResetImageROI(eig)
            # cv.ResetImageROI(temp)
            if not exception_cv:
                for i, fp in enumerate(features):
                    features[i] = (fp[0]+pt1[0]+corrx, fp[1]+pt1[1]+corry)
                exception_cv = False
            need_to_init = False
        # if an average point has been found, draw it as red dot
        if avg:
            cv.Circle (image, (avg[0], avg[1]), 6, (0, 0, 255, 0), -1, 8, 0)
        # if an average running point has been found, draw it as blue dot
        if avg_run:
            cv.Circle (image, (avg_run[0], avg_run[1]), 6, (255, 0, 0, 0), -1, 8, 0)
        if features != []:
            # we have points, so display them
            # calculate the optical flow
            # print "features1", features
            features, status, track_error = cv.CalcOpticalFlowPyrLK (
                prev_grey, grey, prev_pyramid, pyramid,
                features,
                (win_size, win_size), 3,
                (cv.CV_TERMCRIT_ITER|cv.CV_TERMCRIT_EPS, 20, 0.03),
                flags)
            """ from EHCI:
            cvCalcOpticalFlowPyrLK( prev_grey, grey, prev_pyramid, pyramid,
                    points[0], points[1], numberOfTrackingPoints, 
                    cvSize(20,20), 3, status, 0,
                    cvTermCriteria(CV_TERMCRIT_ITER|CV_TERMCRIT_EPS, 20, 0.03), 
                    flags)
            """
            # set back the points we keep
            features = [ p for (st,p) in zip(status, features) if st]
            # print "features15", features, status, track_error
            if add_remove_pt:
                # we have a point to add, so see if it is close to
                # another one. If yes, don't use it
                if min([ ptptdist(pt, p) for p in features ]) < 25:
                    # too close
                    add_remove_pt = 0
            # draw the points as green circles
            minpx = size[0]
            maxpx = 0
            minpy = size[1]
            maxpy = 0
            for the_point in features:
                # draw a green dot for each point in features
                px = int(the_point[0])
                py = int(the_point[1])
                cv.Circle (image, (px, py), 3, (0, 255, 0, 0), -1, 8, 0)
                if px<minpx:
                    minpx = px
                if px>maxpx:
                    maxpx = px
                if py<minpy:
                    minpy = py
                if py>maxpy:
                    maxpy = py
            # distances from min to max in x and y axis of the tracking points
            # at extreme ends
            ddx = maxpx - minpx
            ddy = maxpy - minpy
            if ddy_run == 0: 
                ddy_run = ddy
                # ddy_run_afterpoint = ddy
            # or ddx and ddx_run are not identical and need approach to each other
            else:
                d_y = ddy-ddy_run #_afterpoint
                if d_y != 0:
                    ddy_y_step = 1 # int(abs(d_x/1))
                    # ddy_run_afterpoint = ddy_run_afterpoint + d_y/abs(d_y)*ddy_y_step
                    ddy_run = ddy_run + d_y/abs(d_y)*ddy_y_step
                # ddy_run = int(ddy_run_afterpoint)
            # print "ddy_run", ddy_run
            # also display the ddx and ddy_run as lines at the image border
            # ddy_run in the vertical axis displays the distance from the body
            # to the webcam in a smooth form
            if ddy_run != 0 and ddy != 0:
                cv.Rectangle(image, (2, 2), (ddx*2, 6), cv.RGB(255, 0, 255), 3, 8, 0)
                cv.Rectangle(image, (2, 2), (6, ddy_run*2), cv.RGB(255, 0, 255), 3, 8, 0)
        if add_remove_pt:
            # we want to add a point
            features = add_point(pt, grey, features)
            # we are no longer in "add_remove_pt" mode
            add_remove_pt = False
        # swapping
        prev_grey, grey = grey, prev_grey
        prev_pyramid, pyramid = pyramid, prev_pyramid    
        # we can now display the image
        cv.ShowImage ('LkDemo and Facedetect', image)
        # handle events
        c = cv.WaitKey(10) % 0x100
        if c == 27:
            # user has press the ESC key, so exit
            break
        # processing depending on the character
        if 32 <= c and c < 128:
            cc = chr(c).lower()
            if cc == 'r':
              need_to_init = True
            if cc == 'c':
                features = []
            elif cc == 'n':
                night_mode = not night_mode
        t2 = cv.GetTickCount() - t1
        if t_avg == 0:
            t_avg = t2
            t_all = t2
            tcount = 2
        else:
            t_all = t_all+t2
            t_avg = (t_all)/tcount
            tcount = tcount + 1
        # usually within around 60 ms        
        print "complete round = {0}ms, running average time: {1}ms".format((t2/(cv.GetTickFrequency()*1000.)), (t_avg/(cv.GetTickFrequency()*1000.)))
