#! /usr/bin/env python

print "OpenCV Python version of lkdemo"

import cv, sys, time

# import the necessary things for OpenCV

# image files to import: left01.jpg  left02.jpg  left03.jpg  left04.jpg  left05.jpg  left06.jpg  left07.jpg  left08.jpg  left09.jpg  left11.jpg  left12.jpg  left13.jpg left14.jpg

#############################################################################
# some "constants"

win_size = 10
MAX_COUNT = 500

#############################################################################
# some "global" variables

image = None
pt = None
add_remove_pt = False
flags = 0
night_mode = False
need_to_init = True
CLOCKS_PER_SEC = 1.0
MHI_DURATION = .1
MAX_TIME_DELTA = 0.5
MIN_TIME_DELTA = 0.05
N = 4
buf = range(10) 
last = 0
mhi = None # MHI
orient = None # orientation
mask = None # valid orientation mask
segmask = None # motion segmentation map
storage = cv.CreateMemStorage(0) # temporary storage
diff_threshold = 30
sec_inc = 10
seccount = 0
quality = 0.01
min_distance = 10
move_index = 0
lo_diff = 8
up_diff = 8

#############################################################################
# the mouse callback

# the callback on the trackbar
def on_mouse (event, x, y, flags, param):

    # we will use the global pt and add_remove_pt
    global pt
    global add_remove_pt
    
    if image is None:
        # not initialized, so skip
        return

    if image.origin != 0:
        # different origin
        y = image.height - y

    if event == cv.CV_EVENT_LBUTTONDOWN:
        # user has click, so memorize it
        pt = (x, y)
        add_remove_pt = True
    # print "new point", pt

#############################################################################
# so, here is the main part of the program

if __name__ == '__main__':

    frames = sys.argv[1:]
    if frames == []:
      print "usage lkdemo.py <image files>"
      capture = cv.CaptureFromCAM(0)
      wid = 320
      hei = 240
      ww = wid / 10
      hh = hei / 10
      cv.SetCaptureProperty(capture, cv.CV_CAP_PROP_FRAME_WIDTH, wid)
      cv.SetCaptureProperty(capture, cv.CV_CAP_PROP_FRAME_HEIGHT, hei) 
    else:
      capture = 0
    # display a small howto use it
    print "Hot keys: \n" \
          "\tESC - quit the program\n" \
          "\tr - auto-initialize tracking\n" \
          "\tc - delete all the points\n" \
          "\tn - switch the \"night\" mode on/off\n" \
          "\tSPACE - next frame\n" \
          "To add/remove a feature point click it\n"

    # first, create the necessary windows
    cv.NamedWindow ('LkDemo', cv.CV_WINDOW_AUTOSIZE)
    cv.NamedWindow ('Silhouette', cv.CV_WINDOW_AUTOSIZE)
    cv.NamedWindow ('Result', cv.CV_WINDOW_AUTOSIZE)
    # register the mouse callback
    cv.SetMouseCallback ('LkDemo', on_mouse, None)
    fc = 0
    while 1:
        # do forever
        frame = cv.QueryFrame(capture)
        timestamp = time.clock() / CLOCKS_PER_SEC # get current time in seconds
        idx1 = last
        if image is None:
            # create the images we need
            image = cv.CreateImage (cv.GetSize (frame), 8, 3)
            size = cv.GetSize(image) # get current frame size
            print "size", size
            result = cv.CloneImage(image)
            image.origin = frame.origin
            motion = cv.CreateImage((image.width, image.height), 8, 3)
            cv.Zero(motion)
            # motion.origin = image.origin
            grey = cv.CreateImage (cv.GetSize (image), 8, 1)
            grey2 = cv.CreateImage (cv.GetSize (image), 8, 1)
            grey3 = cv.CreateImage (cv.GetSize (image), 8, 1)
            green = cv.CreateImage (cv.GetSize (image), 8, 3)
            prev_grey = cv.CreateImage (cv.GetSize (image), 8, 1)
            pyramid = cv.CreateImage (cv.GetSize (image), 8, 1)
            prev_pyramid = cv.CreateImage (cv.GetSize (image), 8, 1)
            eig = cv.CreateImage (cv.GetSize (grey2), 32, 1)
            temp = cv.CreateImage (cv.GetSize (grey2), 32, 1)
            features = []
            for i in range(N):
                buf[i] = cv.CreateImage(size, cv.IPL_DEPTH_8U, 1)
                cv.Zero(buf[i])
            mhi = cv.CreateImage(size,cv. IPL_DEPTH_32F, 1)
            cv.Zero(mhi) # clear MHI at the beginning
            orient = cv.CreateImage(size,cv. IPL_DEPTH_32F, 1)
            segmask = cv.CreateImage(size,cv. IPL_DEPTH_32F, 1)
            mask = cv.CreateImage(size,cv. IPL_DEPTH_8U, 1)

        # copy the frame, so we can draw on it
        cv.Copy (frame, image)

        # create a grey version of the image
        # print "image", image.width, image.height
        # print "grey", grey.width, grey.height
        # cv.CvtColor (image, grey, cv.CV_BGR2GRAY)
        # buf[last] = cv.CloneImage(grey)
        # convert frame to grayscale
        cv.CvtColor(image, buf[last], cv.CV_BGR2GRAY) 
        cv.CvtColor (image, grey2, cv.CV_BGR2GRAY)
        idx2 = (last + 1) % N # index of (last - (N-1))th frame
        last = idx2
        silh = buf[idx2]
        cv.AbsDiff(buf[idx1], buf[idx2], silh) # get difference between frames
        cv.Threshold(silh, silh, diff_threshold, 1, cv.CV_THRESH_BINARY) # and threshold it
        contours = cv.FindContours(silh, storage, cv.CV_RETR_TREE,\
        cv.CV_CHAIN_APPROX_SIMPLE, (0,0))
        if contours:
            cv.DrawContours(silh, contours, cv.ScalarAll(255), cv.ScalarAll(255),\
            1, 1, cv.CV_AA)
        red = cv.CloneImage(silh)
        blue = cv.CloneImage(silh)
        cv.Dilate(silh, silh, None, 8)
        # seed = (wid/2, hei/2)
        # cv.Circle (blue, seed, 3, (255, 0, 0, 0), -1, 8, 0)        
        # cv.FloodFill(silh, seed, cv.RGB(0,255,0), cv.CV_RGB(lo_diff, lo_diff, lo_diff),\
        # cv.CV_RGB(up_diff, up_diff, up_diff), 8, None)
        cv.Erode(silh, silh, None, 8)
        cv.Merge(blue, silh, red, None, green)
        cv.ShowImage("Silhouette", green)
        # old_ grey = cv.CloneImage()
        cv.CvtColor(green, grey, cv.CV_BGR2GRAY) 
        moving = cv.Avg(grey)[0]
        move_index += sec_inc
        # print moving # indicates according to value the moment of movement in the image
        
        if night_mode:
            # night mode: only display the points
            cv.SetZero(image)

        if need_to_init or (moving>.1 and move_index>1000) or len(features)<20:
            print "retracked",
            move_index = 0
            # we want to search all the good points

            # search the good points
            features_new = cv.GoodFeaturesToTrack (
                grey, eig, temp,
                MAX_COUNT,
                quality, min_distance, None, 3, 0, 0.04)
            # print features
            if len(features_new)>20:
                features = features_new[:]
                # refine the corner locations
                features = cv.FindCornerSubPix (
                    grey2,
                    features,
                    (win_size, win_size),  (-1, -1),
                    (cv.CV_TERMCRIT_ITER | cv.CV_TERMCRIT_EPS, 20, 0.03))
            else:
                pass

        elif features != []:
            # we have points, so display them

            # calculate the optical flow
            features, status, track_error = cv.CalcOpticalFlowPyrLK (
                prev_grey, grey2, prev_pyramid, pyramid,
                features,
                (win_size, win_size), 3,
                (cv.CV_TERMCRIT_ITER|cv.CV_TERMCRIT_EPS, 20, 0.03),
                flags)

            # set back the points we keep
            features = [ p for (st,p) in zip(status, features) if st]

            if add_remove_pt:
                # we have a point to add, so see if it is close to
                # another one. If yes, don't use it
                def ptptdist(p0, p1):
                    dx = p0[0] - p1[0]
                    dy = p0[1] - p1[1]
                    return dx**2 + dy**2
                if min([ ptptdist(pt, p) for p in features ]) < 25:
                    # too close
                    add_remove_pt = 0

            # draw the points as green circles
            cv.Zero(result)
            old_point = None
            a = len(features)
            avg = [0, 0]
            avg2 = [0, 0]
            old_features = features[:]
            features1 = []
            features2 = []
            for the_point in features:
                y, x = int(the_point[1]), int(the_point[0])
                if old_point != None:
                    if 0<=y<size[1] and 0<=x<size[0]:
                        point = cv.Get2D(green, y, x)[1]
                        if point != 0:
                            cv.Circle (result, (x, y), 3, (0, 0, 255, 0), -1, 8, 0)
                            features1.append((y, x))
                            avg = [avg[0]+x, avg[1]+y]
                        else:
                            cv.Circle (result, (x, y), 3, (255, 0, 0, 0), -1, 8, 0)
                            features2.append((y, x))
                            avg2 = [avg2[0]+x, avg2[1]+y]
                    # new_features.append(the_point)
                old_point = the_point
            # print "new_features", new_features
            if len(features1)>5:
                features3 = features1[:]
            elif len(features2)>5:
                features3 = features2[:]
            hull = cv.ConvexHull2(features3, storage, cv.CV_CLOCKWISE, 1)
            # cv.FillPoly(result, [hull], cv.RGB(0,255,0), cv.CV_AA)
            cv.PolyLine(result, [hull], 1, cv.RGB(255,0,0), 1, cv.CV_AA)
            a = len(features3)
            a2 = len(features2)
            if a2 != 0:
                avg2 = [int(avg2[0]/a2), int(avg2[1]/a2)]
            if a != 0:
                avg = [int(avg[0]/a), int(avg[1]/a)]
            else:
                avg = avg2
            # print avg, 
            # cv.Circle (result, (avg[0], avg[1]), ww+((hei-avg[1])/4), (255, 255, 255, 0), -1, 8, 0)
            # cv.Circle (image, (avg[0], avg[1]), ww+((hei-avg[1])/4), (255, 0, 255, 0), -1, 8, 0)
            cv.Circle (result, (avg[0], avg[1]), wid/32, (0, 0, 255, 0), -1, 8, 0)
            cv.Circle (result, (avg2[0], avg2[1]), wid/32, (255, 0, 0, 0), -1, 8, 0)
            # cv.Circle (image, (avg[0], avg[1]), 6, (255, 0, 0, 0), -1, 8, 0)
            # cv.And(image, result, image)
            cv.Circle (image, (avg[0], avg[1]), wid/32, (0, 0, 255, 0), -1, 8, 0)
            # cv.EllipseBox(result, ((float(avg[0]-ww), float(avg[1]-hh*2)),\
            # (float(avg[0]+ww), float(avg[1]+hh*2)), 10.0), (255, 0, 255, 0), 1, 8)
            # (avg[0]-ww, avg[1]-hh*2, avg[0]+ww, avg[1]+hh*2)
                
        if add_remove_pt:
            # we want to add a point
            # refine this corner location and append it to 'features'

            features += cv.FindCornerSubPix (
                grey2,
                [pt],
                (win_size, win_size),  (-1, -1),
                (cv.CV_TERMCRIT_ITER | cv.CV_TERMCRIT_EPS, 20, 0.03))
            # we are no longer in "add_remove_pt" mode
            add_remove_pt = False

        # swapping
        prev_grey, grey2 = grey2, prev_grey
        prev_pyramid, pyramid = pyramid, prev_pyramid
        need_to_init = False
        
        # we can now display the image
        cv.ShowImage ('LkDemo', image)
        cv.ShowImage ('Result', result)
        # handle events
        c = cv.WaitKey(sec_inc) % 0x100
        seccount += sec_inc
        if seccount % 1000 == 0:
            seccount = 0
            need_to_init = True

        if c == 27:
            # user has press the ESC key, so exit
            break

        # processing depending on the character
        if 32 <= c and c < 128:
          cc = chr(c).lower()
          if cc == 'r':
              need_to_init = True
          elif cc == 'c':
              features = []
          elif cc == 'n':
              night_mode = not night_mode
          elif cc == ' ':
            try:
              fc = (fc + 1) % len(frames)
            except ZeroDivisionError:
              pass
