import cv2.cv as cv

def skin_detect(image, show_results=False):
    # original image
    if show_results:
        cv.NamedWindow("Input", 1)
        cv.ShowImage("Input", image)

    # create hsv image
    hsv = cv.CreateImage(cv.GetSize(image), 8, 3)
    cv.Zero(hsv)
    cv.CvtColor(image, hsv, cv.CV_BGR2HSV)
    # show it
    if show_results:
        cv.NamedWindow("HSV", 1)
        cv.ShowImage("HSV", hsv)

    # split the hsv image
    h = cv.CreateImage(cv.GetSize(hsv), 8, 1)
    s = cv.CreateImage(cv.GetSize(hsv), 8, 1)
    v = cv.CreateImage(cv.GetSize(hsv), 8, 1)
    cv.Split(hsv, h, s, v, None)

    t = cv.CreateImage(cv.GetSize(hsv), 8, 1)
    dst = cv.CreateImage(cv.GetSize(hsv), 8, 1)
    cv.Zero(dst)

    if show_results:
        cv.NamedWindow("Hue", 1)
        cv.ShowImage("Hue", h)
        cv.NamedWindow("Saturation", 1)
        cv.ShowImage("Saturation", s)
        cv.NamedWindow("Value", 1)
        cv.ShowImage("Value", v)

    # Detect which pixels in each of the H, S and V channels are probably skin pixels.
    # Assume that skin has a Hue between 0 to 18 (out of 180), 
    # and Saturation above 50, and Brightness abov
    cv.Threshold(h, h, 18, 255, cv.CV_THRESH_BINARY_INV)
    cv.Threshold(s, s, 50, 255, cv.CV_THRESH_BINARY)
    cv.Threshold(v, v, 80, 255, cv.CV_THRESH_BINARY)

    if show_results:
        cv.NamedWindow("H Treshold", 1)
        cv.ShowImage("H Treshold", h)
        cv.NamedWindow("S Treshold", 1)
        cv.ShowImage("S Treshold", s)
        cv.NamedWindow("V Treshold", 1)
        cv.ShowImage("V Treshold", v)
    # Greyscale output image
    skin = cv.CreateImage(cv.GetSize(hsv), 8, 1)
    # imageSkin = H {BITWISE_AND} S
    cv.And(h, s, skin)
    cv.And(skin, v, skin)
    if show_results:
        cv.NamedWindow("Skin", 1)
        cv.ShowImage("Skin", skin)
    return skin

if __name__ == '__main__':
    import sys, time
    if len(sys.argv)<2:
        image = cv.LoadImage("Skin.jpg")
    else:
        image = cv.LoadImage(sys.argv[1])    
    
    show_results = True
    tim = time.time()
    skin = skin_detect(image, show_results)
    print "Skin detection in", time.time()-tim, "secs"
    cv.WaitKey()